@@ -3,7 +3,7 @@
  */
 
 /*
- * Copyright (C) 2012-2018 the Free Software Foundation, Inc.
+ * Copyright (C) 2012-2019, 2021, 2022, the Free Software Foundation, Inc.
  *
  * This file is part of GAWK, the GNU implementation of the
  * AWK Programming Language.
@@ -25,21 +25,14 @@
 
 #include "awk.h"
 
-#ifdef HAVE_MPFR
-#define getmpfr(n)	getblock(n, BLOCK_MPFR, mpfr_ptr)
-#define freempfr(n)	freeblock(n, BLOCK_MPFR)
-
-#define getmpz(n)	getblock(n, BLOCK_MPZ, mpz_ptr)
-#define freempz(n)	freeblock(n, BLOCK_MPZ)
-#endif
-
 /* Declare some globals used by api_get_file: */
 extern IOBUF *curfile;
 extern INSTRUCTION *main_beginfile;
 extern int currule;
 
 static awk_bool_t node_to_awk_value(NODE *node, awk_value_t *result, awk_valtype_t wanted);
-static char *valtype2str(awk_valtype_t type);
+static const char *valtype2str(awk_valtype_t type);
+static NODE *ns_lookup(const char *name_space, const char *name, char **full_name);
 
 /*
  * api_get_argument --- get the count'th paramater, zero-based.
@@ -76,9 +69,9 @@ api_get_argument(awk_ext_id_t id, size_t count,
 		return awk_false;
 
 	/* if type is undefined */
-	if (arg->type == Node_var_new) {
+	if (arg->type == Node_var_new || arg->type == Node_elem_new) {
 		if (wanted == AWK_UNDEFINED)
-			return true;
+			return awk_true;
 		else if (wanted == AWK_ARRAY) {
 			goto array;
 		} else {
@@ -89,7 +82,7 @@ api_get_argument(awk_ext_id_t id, size_t count,
 	/* at this point, we have real type */
 	if (arg->type == Node_var_array || arg->type == Node_array_ref) {
 		if (wanted != AWK_ARRAY && wanted != AWK_UNDEFINED)
-			return false;
+			return awk_false;
 		goto array;
 	} else
 		goto scalar;
@@ -131,7 +124,7 @@ api_set_argument(awk_ext_id_t id,
 		return awk_false;
 
 	if (   (arg = get_argument(count)) == NULL
-	    || arg->type != Node_var_new)
+	    || (arg->type != Node_var_new && arg->type != Node_elem_new))
 		return awk_false;
 
 	arg = get_array_argument(arg, count);
@@ -155,6 +148,9 @@ awk_value_to_node(const awk_value_t *retval)
 {
 	NODE *ext_ret_val = NULL;
 	NODE *v;
+#ifdef HAVE_MPFR
+	int tval = 0;
+#endif
 
 	if (retval == NULL)
 		fatal(_("awk_value_to_node: received null retval"));
@@ -166,6 +162,9 @@ awk_value_to_node(const awk_value_t *retval)
 	case AWK_UNDEFINED:
 		ext_ret_val = dupnode(Nnull_string);
 		break;
+	case AWK_BOOL:
+		ext_ret_val = make_bool_node(retval->bool_value != awk_false);
+		break;
 	case AWK_NUMBER:
 		switch (retval->num_type) {
 		case AWK_NUMBER_TYPE_DOUBLE:
@@ -176,8 +175,10 @@ awk_value_to_node(const awk_value_t *retval)
 			if (! do_mpfr)
 				fatal(_("awk_value_to_node: not in MPFR mode"));
 			ext_ret_val = make_number_node(MPFN);
-			memcpy(&ext_ret_val->mpg_numbr, retval->num_ptr, sizeof(ext_ret_val->mpg_numbr));
-			freempfr(retval->num_ptr);
+			mpfr_init(ext_ret_val->mpg_numbr);
+			tval = mpfr_set(ext_ret_val->mpg_numbr, (mpfr_srcptr) retval->num_ptr, ROUND_MODE);
+			IEEE_FMT(ext_ret_val->mpg_numbr, tval);
+			mpfr_clear(retval->num_ptr);
 #else
 			fatal(_("awk_value_to_node: MPFR not supported"));
 #endif
@@ -187,8 +188,9 @@ awk_value_to_node(const awk_value_t *retval)
 			if (! do_mpfr)
 				fatal(_("awk_value_to_node: not in MPFR mode"));
 			ext_ret_val = make_number_node(MPZN);
-			memcpy(&ext_ret_val->mpg_i, retval->num_ptr, sizeof(ext_ret_val->mpg_i));
-			freempz(retval->num_ptr);
+			mpz_init(ext_ret_val->mpg_i);
+			mpz_set(ext_ret_val->mpg_i, (mpz_ptr) retval->num_ptr);
+			mpz_clear(retval->num_ptr);
 #else
 			fatal(_("awk_value_to_node: MPFR not supported"));
 #endif
@@ -283,7 +285,7 @@ api_lintwarn(awk_ext_id_t id, const char *format, ...)
 	(void) id;
 
 	va_start(args, format);
-	if (lintwarn == r_fatal) {
+	if (lintfunc == r_fatal) {
 		err(true, _("fatal: "), format, args);
 	} else {
 		err(false, _("warning: "), format, args);
@@ -372,17 +374,19 @@ api_unset_ERRNO(awk_ext_id_t id)
 
 static awk_bool_t
 api_add_ext_func(awk_ext_id_t id,
-		const char *namespace,
+		const char *name_space,
 		awk_ext_func_t *func)
 {
 	(void) id;
-	(void) namespace;
 
 	if (func == NULL)
 		return awk_false;
 
+	if (name_space == NULL)
+		fatal(_("add_ext_func: received NULL name_space parameter"));
+
 #ifdef DYNAMIC
-	return make_builtin(func);
+	return make_builtin(name_space, func);
 #else
 	return awk_false;
 #endif
@@ -517,7 +521,7 @@ assign_number(NODE *node, awk_value_t *val)
 		val->num_ptr = &node->mpg_i;
 		break;
 	default:
-		fatal(_("node_to_awk_value: detected invalid numeric flags combination `%s'; please file a bug report."), flags2str(node->flags));
+		fatal(_("node_to_awk_value: detected invalid numeric flags combination `%s'; please file a bug report"), flags2str(node->flags));
 		break;
 	}
 #endif
@@ -537,6 +541,16 @@ assign_regex(NODE *node, awk_value_t *val)
 	val->val_type = AWK_REGEX;
 }
 
+/* assign_bool --- return a bool node */
+
+static inline void
+assign_bool(NODE *node, awk_value_t *val)
+{
+	assert((node->flags & BOOLVAL) != 0);
+	val->val_type = AWK_BOOL;
+	val->bool_value = get_number_si(node) != 0 ? awk_true : awk_false;
+}
+
 /* node_to_awk_value --- convert a node into a value for an extension */
 
 static awk_bool_t
@@ -552,6 +566,7 @@ node_to_awk_value(NODE *node, awk_value_t *val, awk_valtype_t wanted)
 
 	switch (node->type) {
 	case Node_var_new:	/* undefined variable */
+	case Node_elem_new:	/* undefined element */
 		val->val_type = AWK_UNDEFINED;
 		if (wanted == AWK_UNDEFINED) {
 			ret = awk_true;
@@ -572,8 +587,16 @@ node_to_awk_value(NODE *node, awk_value_t *val, awk_valtype_t wanted)
 	case Node_val:
 		/* a scalar value */
 		switch (wanted) {
+		case AWK_BOOL:
+			if ((node->flags & BOOLVAL) != 0) {
+				assign_bool(node, val);
+				ret = awk_true;
+			} else
+				ret = awk_false;
+			break;
+
 		case AWK_NUMBER:
-			if (node->flags & REGEX)
+			if ((node->flags & REGEX) != 0)
 				val->val_type = AWK_REGEX;
 			else {
 				(void) force_number(node);
@@ -583,7 +606,10 @@ node_to_awk_value(NODE *node, awk_value_t *val, awk_valtype_t wanted)
 			break;
 
 		case AWK_STRNUM:
-			switch (fixtype(node)->flags & (STRING|NUMBER|USER_INPUT|REGEX)) {
+			switch (fixtype(node)->flags & (STRING|NUMBER|USER_INPUT|REGEX|BOOLVAL)) {
+			case NUMBER|BOOLVAL:
+				val->val_type = AWK_BOOL;
+				break;
 			case STRING:
 				val->val_type = AWK_STRING;
 				break;
@@ -604,7 +630,7 @@ node_to_awk_value(NODE *node, awk_value_t *val, awk_valtype_t wanted)
 				}
 				/* fall through */
 			default:
-				warning(_("node_to_awk_value detected invalid flags combination `%s'; please file a bug report."), flags2str(node->flags));
+				warning(_("node_to_awk_value detected invalid flags combination `%s'; please file a bug report"), flags2str(node->flags));
 				val->val_type = AWK_UNDEFINED;
 				break;
 			}
@@ -617,10 +643,13 @@ node_to_awk_value(NODE *node, awk_value_t *val, awk_valtype_t wanted)
 			break;
 
 		case AWK_REGEX:
-			switch (fixtype(node)->flags & (STRING|NUMBER|USER_INPUT|REGEX)) {
+			switch (fixtype(node)->flags & (STRING|NUMBER|USER_INPUT|REGEX|BOOLVAL)) {
 			case STRING:
 				val->val_type = AWK_STRING;
 				break;
+			case NUMBER|BOOLVAL:
+				val->val_type = AWK_BOOL;
+				break;
 			case NUMBER:
 				val->val_type = AWK_NUMBER;
 				break;
@@ -638,14 +667,17 @@ node_to_awk_value(NODE *node, awk_value_t *val, awk_valtype_t wanted)
 				}
 				/* fall through */
 			default:
-				warning(_("node_to_awk_value detected invalid flags combination `%s'; please file a bug report."), flags2str(node->flags));
+				warning(_("node_to_awk_value detected invalid flags combination `%s'; please file a bug report"), flags2str(node->flags));
 				val->val_type = AWK_UNDEFINED;
 				break;
 			}
 			break;
 
 		case AWK_SCALAR:
-			switch (fixtype(node)->flags & (STRING|NUMBER|USER_INPUT|REGEX)) {
+			switch (fixtype(node)->flags & (STRING|NUMBER|USER_INPUT|REGEX|BOOLVAL)) {
+			case NUMBER|BOOLVAL:
+				val->val_type = AWK_BOOL;
+				break;
 			case STRING:
 				val->val_type = AWK_STRING;
 				break;
@@ -665,7 +697,7 @@ node_to_awk_value(NODE *node, awk_value_t *val, awk_valtype_t wanted)
 				}
 				/* fall through */
 			default:
-				warning(_("node_to_awk_value detected invalid flags combination `%s'; please file a bug report."), flags2str(node->flags));
+				warning(_("node_to_awk_value detected invalid flags combination `%s'; please file a bug report"), flags2str(node->flags));
 				val->val_type = AWK_UNDEFINED;
 				break;
 			}
@@ -673,7 +705,11 @@ node_to_awk_value(NODE *node, awk_value_t *val, awk_valtype_t wanted)
 
 		case AWK_UNDEFINED:
 			/* return true and actual type for request of undefined */
-			switch (fixtype(node)->flags & (STRING|NUMBER|USER_INPUT|REGEX)) {
+			switch (fixtype(node)->flags & (STRING|NUMBER|USER_INPUT|REGEX|BOOLVAL)) {
+			case NUMBER|BOOLVAL:
+				assign_bool(node, val);
+				ret = awk_true;
+				break;
 			case STRING:
 				assign_string(node, val, AWK_STRING);
 				ret = awk_true;
@@ -698,7 +734,7 @@ node_to_awk_value(NODE *node, awk_value_t *val, awk_valtype_t wanted)
 				}
 				/* fall through */
 			default:
-				warning(_("node_to_awk_value detected invalid flags combination `%s'; please file a bug report."), flags2str(node->flags));
+				warning(_("node_to_awk_value detected invalid flags combination `%s'; please file a bug report"), flags2str(node->flags));
 				val->val_type = AWK_UNDEFINED;
 				break;
 			}
@@ -747,6 +783,7 @@ node_to_awk_value(NODE *node, awk_value_t *val, awk_valtype_t wanted)
 
 static awk_bool_t
 api_sym_lookup(awk_ext_id_t id,
+		const char *name_space,
 		const char *name,
 		awk_valtype_t wanted,
 		awk_value_t *result)
@@ -758,7 +795,12 @@ api_sym_lookup(awk_ext_id_t id,
 	if (   name == NULL
 	    || *name == '\0'
 	    || result == NULL
-	    || (node = lookup(name)) == NULL)
+	    || ! is_valid_identifier(name)
+	    || name_space == NULL
+	    || (name_space[0] != '\0' && ! is_valid_identifier(name_space)))
+		return awk_false;
+
+	if ((node = ns_lookup(name_space, name, NULL)) == NULL)
 		return awk_false;
 
 	if (is_off_limits_var(name))	/* a built-in variable */
@@ -791,6 +833,7 @@ api_sym_lookup_scalar(awk_ext_id_t id,
 
 static awk_bool_t
 api_sym_update(awk_ext_id_t id,
+		const char *name_space,
 		const char *name,
 		awk_value_t *value)
 {
@@ -799,7 +842,10 @@ api_sym_update(awk_ext_id_t id,
 
 	if (   name == NULL
 	    || *name == '\0'
-	    || value == NULL)
+	    || value == NULL
+	    || ! is_valid_identifier(name)
+	    || name_space == NULL
+	    || (name_space[0] != '\0' && ! is_valid_identifier(name_space)))
 		return awk_false;
 
 	switch (value->val_type) {
@@ -818,22 +864,21 @@ api_sym_update(awk_ext_id_t id,
 		return awk_false;
 	}
 
-	node = lookup(name);
+	char *full_name = NULL;
+	node = ns_lookup(name_space, name, & full_name);
 
 	if (node == NULL) {
 		/* new value to be installed */
 		if (value->val_type == AWK_ARRAY) {
 			array_node = awk_value_to_node(value);
-			node = install_symbol(estrdup((char *) name, strlen(name)),
-					Node_var_array);
+			node = install_symbol(full_name, Node_var_array);
 			array_node->vname = node->vname;
 			*node = *array_node;
 			freenode(array_node);
 			value->array_cookie = node;	/* pass new cookie back to extension */
 		} else {
 			/* regular variable */
-			node = install_symbol(estrdup((char *) name, strlen(name)),
-					Node_var);
+			node = install_symbol(full_name, Node_var);
 			node->var_value = awk_value_to_node(value);
 		}
 
@@ -842,19 +887,25 @@ api_sym_update(awk_ext_id_t id,
 
 	/*
 	 * If we get here, then it exists already.  Any valid type is
-	 * OK except for AWK_ARRAY.
+	 * OK except for AWK_ARRAY (unless it is in Node_var_new undefined
+	 * state, in which case an array is OK).
 	 */
 	if (   (node->flags & NO_EXT_SET) != 0
-	    || is_off_limits_var(name)) {	/* most built-in vars not allowed */
+	    || is_off_limits_var(full_name)) {	/* most built-in vars not allowed */
 		node->flags |= NO_EXT_SET;
+		efree((void *) full_name);
 		return awk_false;
 	}
 
-	if (    value->val_type != AWK_ARRAY
-	    && (node->type == Node_var || node->type == Node_var_new)) {
+	efree((void *) full_name);
+
+	if ((node->type == Node_var && value->val_type != AWK_ARRAY)
+	    || node->type == Node_var_new
+	    || node->type == Node_elem_new) {
 		unref(node->var_value);
 		node->var_value = awk_value_to_node(value);
-		if (node->type == Node_var_new && value->val_type != AWK_UNDEFINED)
+		if ((node->type == Node_var_new || node->type == Node_elem_new)
+		    && value->val_type != AWK_UNDEFINED)
 			node->type = Node_var;
 
 		return awk_true;
@@ -1031,7 +1082,6 @@ api_set_array_element(awk_ext_id_t id, awk_array_t a_cookie,
 	NODE *array = (NODE *)a_cookie;
 	NODE *tmp;
 	NODE *elem;
-	NODE **aptr;
 
 	/* don't check for index len zero, null str is ok as index */
 	if (   array == NULL
@@ -1043,17 +1093,13 @@ api_set_array_element(awk_ext_id_t id, awk_array_t a_cookie,
 		return awk_false;
 
 	tmp = awk_value_to_node(index);
-	aptr = assoc_lookup(array, tmp);
-	unref(*aptr);
-	elem = *aptr = awk_value_to_node(value);
+	elem = awk_value_to_node(value);
 	if (elem->type == Node_var_array) {
 		elem->parent_array = array;
 		elem->vname = estrdup(index->str_value.str,
 					index->str_value.len);
 	}
-	if (array->astore != NULL)
-		(*array->astore)(array, tmp);
-	unref(tmp);
+	assoc_set(array, tmp, elem);
 
 	return awk_true;
 }
@@ -1164,6 +1210,17 @@ api_clear_array(awk_ext_id_t id, awk_array_t a_cookie)
 	return awk_true;
 }
 
+/* api_destroy_array --- destroy an array */
+
+static awk_bool_t
+api_destroy_array(awk_ext_id_t id, awk_array_t a_cookie)
+{
+	if (! api_clear_array(id, a_cookie))
+		return awk_false;
+	freenode((NODE *) a_cookie);
+	return awk_true;
+}
+
 /* api_flatten_array_typed --- flatten out an array so that it can be looped over easily. */
 
 static awk_bool_t
@@ -1179,7 +1236,7 @@ api_flatten_array_typed(awk_ext_id_t id,
 
 	if (   array == NULL
 	    || array->type != Node_var_array
-	    || array->table_size == 0
+	    || assoc_empty(array)
 	    || data == NULL)
 		return awk_false;
 
@@ -1204,12 +1261,12 @@ api_flatten_array_typed(awk_ext_id_t id,
 		/* Convert index and value to API types. */
 		if (! node_to_awk_value(index,
 				& (*data)->elements[j].index, index_type)) {
-			fatal(_("api_flatten_array_typed: could not convert index %d to %s\n"),
+			fatal(_("api_flatten_array_typed: could not convert index %d to %s"),
 						(int) i, valtype2str(index_type));
 		}
 		if (! node_to_awk_value(value,
 				& (*data)->elements[j].value, value_type)) {
-			fatal(_("api_flatten_array_typed: could not convert value %d to %s\n"),
+			fatal(_("api_flatten_array_typed: could not convert value %d to %s"),
 						(int) i, valtype2str(value_type));
 		}
 	}
@@ -1227,7 +1284,7 @@ api_release_flattened_array(awk_ext_id_t id,
 		awk_array_t a_cookie,
 		awk_flat_array_t *data)
 {
-	NODE *array = a_cookie;
+	NODE *array = (NODE *) a_cookie;
 	NODE **list;
 	size_t i, j, k;
 
@@ -1301,7 +1358,7 @@ api_get_mpfr(awk_ext_id_t id)
 {
 #ifdef HAVE_MPFR
 	mpfr_ptr p;
-	getmpfr(p);
+	emalloc(p, mpfr_ptr, sizeof(mpfr_t), "api_get_mpfr");
 	mpfr_init(p);
 	return p;
 #else
@@ -1317,7 +1374,8 @@ api_get_mpz(awk_ext_id_t id)
 {
 #ifdef HAVE_MPFR
 	mpz_ptr p;
-	getmpz(p);
+	emalloc(p, mpz_ptr, sizeof (mpz_t), "api_get_mpz");
+
 	mpz_init(p);
 	return p;
 #else
@@ -1520,6 +1578,9 @@ gawk_api_t api_impl = {
 
 	/* Find/open a file */
 	api_get_file,
+
+	/* Additional array hook to destroy an array */
+	api_destroy_array,
 };
 
 /* init_ext_api --- init the extension API */
@@ -1557,13 +1618,13 @@ print_ext_versions(void)
 
 /* valtype2str --- return a printable representation of a value type */
 
-static char *
+static const char *
 valtype2str(awk_valtype_t type)
 {
 	static char buf[100];
 
 	// Important: keep in same order as in gawkapi.h!
-	static char *values[] = {
+	static const char *values[] = {
 		"AWK_UNDEFINED",
 		"AWK_NUMBER",
 		"AWK_STRING",
@@ -1581,3 +1642,31 @@ valtype2str(awk_valtype_t type)
 
 	return buf;
 }
+
+/* ns_lookup --- correctly build name before looking it up */
+
+static NODE *
+ns_lookup(const char *name_space, const char *name, char **fullname)
+{
+	assert(name_space != NULL);
+	assert(name != NULL);
+
+	if (name_space[0] == '\0' || strcmp(name_space, awk_namespace) == 0) {
+		if (fullname != NULL)
+			*fullname = estrdup(name, strlen(name));
+		return lookup(name);
+	}
+
+	size_t len = strlen(name_space) + 2 + strlen(name) + 1;
+	char *buf;
+	emalloc(buf, char *, len, "ns_lookup");
+	sprintf(buf, "%s::%s", name_space, name);
+
+	NODE *f = lookup(buf);
+	if (fullname != NULL)
+		*fullname = buf;
+	else
+		efree((void *) buf);
+
+	return f;
+}
